/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef _H_JFS_LOCK
#define _H_JFS_LOCK

#ifdef kern22
#include <asm/spinlock.h>
#else
#include <linux/spinlock.h>
#endif
#include <linux/sched.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,18)
typedef struct wait_queue *wait_queue_head_t;
#define init_waitqueue_head init_waitqueue
#define init_MUTEX(sem) *(sem) = MUTEX
#define DECLARE_MUTEX_LOCKED(sem) struct semaphore sem = MUTEX_LOCKED
#define DECLARE_WAITQUEUE(wait, current)\
	struct wait_queue wait = { current, NULL }
#endif

/*
 *	jfs_lock.h
 *
 * JFS lock definition for globally referenced locks
 */

/*
 * General purpose spinlock for really quick stuff
 */
extern spinlock_t jfs_spinlock;

#define	MUTEXLOCK_T	struct semaphore
#define MUTEXLOCK_ALLOC(Lock,Flag,Class,Occurrence)
#define MUTEXLOCK_FREE(Lock)
#ifdef kern22
#define MUTEXLOCK_INIT(Lock) *(Lock) = MUTEX
#else
#define MUTEXLOCK_INIT(Lock) init_MUTEX(Lock)
#endif
#define MUTEXLOCK_LOCK(Lock) down(Lock)
#define MUTEXLOCK_UNLOCK(Lock) up(Lock)
#define MUTEXLOCK_NOLOCK(Lock)

/* readers/writer lock: thread-thread */
/*
 * see jfs_txnmgr.h.  Since jfs_lock.h is included first, we need to define
 * IDIRTY here
 */
#define IDIRTY	0x0080		/* Tried to write dirty inode while locked */

#ifdef kern22
#define JFS_EXCLUSIVE_LOCKED	1
extern spinlock_t jfs_rwlock_lock;

#define	RDWRLOCK_T	jfs_rwlock_t
#define RDWRLOCK_ALLOC(Lock,Flag,Class,Occurrence)
#define RDWRLOCK_FREE(Lock)
#define RDWRLOCK_INIT(Lock)\
{\
	init_waitqueue_head(&(Lock)->wq);\
	(Lock)->flag = (Lock)->readers = (Lock)->writers = 0;\
}

static inline void READ_LOCK(jfs_rwlock_t * Lock)
{
	DECLARE_WAITQUEUE(wait, current);

	spin_lock(&jfs_rwlock_lock);
	if ((Lock->flag & JFS_EXCLUSIVE_LOCKED) ||
	    (Lock->readers && Lock->writers)) {
		add_wait_queue(&Lock->wq, &wait);
		while ((Lock->flag & JFS_EXCLUSIVE_LOCKED) ||
		       (Lock->readers && Lock->writers)) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			spin_unlock(&jfs_rwlock_lock);
			schedule();
			spin_lock(&jfs_rwlock_lock);
		}
		remove_wait_queue(&Lock->wq, &wait);
	}
	Lock->readers++;
	spin_unlock(&jfs_rwlock_lock);
}

static inline void READ_UNLOCK(struct inode *ip, jfs_rwlock_t * Lock)
{
	uint nreaders;

	spin_lock(&jfs_rwlock_lock);
	nreaders = --(Lock)->readers;
	spin_unlock(&jfs_rwlock_lock);
	if (nreaders == 0) {
		wake_up(&Lock->wq);
		if (ip->i_jfs_flag & IDIRTY) {
			mark_inode_dirty(ip);
			ip->i_jfs_flag &= ~IDIRTY;
		}
	}
}

static inline void WRITE_LOCK(jfs_rwlock_t * Lock)
{
	DECLARE_WAITQUEUE(wait, current);

	spin_lock(&jfs_rwlock_lock);
	if ((Lock->flag & JFS_EXCLUSIVE_LOCKED) || Lock->readers) {
		Lock->writers++;
		add_wait_queue(&Lock->wq, &wait);
		while ((Lock->flag & JFS_EXCLUSIVE_LOCKED)
		       || Lock->readers) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			spin_unlock(&jfs_rwlock_lock);
			schedule();
			spin_lock(&jfs_rwlock_lock);
		}
		remove_wait_queue(&Lock->wq, &wait);
		Lock->writers--;
	}
	Lock->flag |= JFS_EXCLUSIVE_LOCKED;
	spin_unlock(&jfs_rwlock_lock);
}

static inline int WRITE_TRYLOCK(jfs_rwlock_t * Lock)
{
	spin_lock(&jfs_rwlock_lock);
	if ((Lock->flag & JFS_EXCLUSIVE_LOCKED) || Lock->readers) {
		spin_unlock(&jfs_rwlock_lock);
		return 0;
	}
	Lock->flag |= JFS_EXCLUSIVE_LOCKED;
	spin_unlock(&jfs_rwlock_lock);
	return 1;
}

static inline void WRITE_UNLOCK(struct inode * ip, jfs_rwlock_t * Lock)
{
	Lock->flag &= ~JFS_EXCLUSIVE_LOCKED;
	wake_up(&Lock->wq);
	if (ip->i_jfs_flag & IDIRTY) {
		mark_inode_dirty(ip);
		ip->i_jfs_flag &= ~IDIRTY;
	}
}
#else				/* ! kern22 ( >= 2.4.0 )  */

/*
 * RW semaphores do not currently have a trylock function.  Since the
 * implementation varies by platform, I have implemented a platform-independent
 * wrapper around the rw_semaphore routines.  If this turns out to be the best
 * way of avoiding our locking problems, I will push to get a trylock
 * implemented in the kernel.
 */
#define RDWRLOCK_T jfs_rwlock_t
#define RDWRLOCK_ALLOC(Lock,Flag,Class,Occurrence)
#define RDWRLOCK_FREE(Lock)
static inline void RDWRLOCK_INIT(jfs_rwlock_t *Lock)
{
	init_rwsem(&Lock->rw_sem);
	atomic_set(&Lock->in_use, 0);
}
static inline void READ_LOCK(jfs_rwlock_t *Lock)
{
	down_read(&Lock->rw_sem);
	atomic_inc(&Lock->in_use);
}
static inline void READ_UNLOCK(struct inode *ip, jfs_rwlock_t *Lock)
{
	atomic_dec(&Lock->in_use);
	up_read(&Lock->rw_sem);
	if (atomic_read(&Lock->in_use) && (ip->i_jfs_flag & IDIRTY)) {
		mark_inode_dirty(ip);
		ip->i_jfs_flag &= ~IDIRTY;
	}
}
static inline void WRITE_LOCK(jfs_rwlock_t *Lock)
{
	down_write(&Lock->rw_sem);
	atomic_inc(&Lock->in_use);
}
/*
 * I'm not worried about a race condition between processors.  The problem
 * I'm avoiding is trying to take a nested lock when memory pressure forces
 * the kernel to try to write all the dirty inodes.
 */
static inline int WRITE_TRYLOCK(jfs_rwlock_t *Lock)
{
	if (atomic_read(&Lock->in_use))
		return 0;
	WRITE_LOCK(Lock);
	return 1;
}
static inline void WRITE_UNLOCK(struct inode *ip, jfs_rwlock_t *Lock)
{
	atomic_dec(&Lock->in_use);
	up_write(&Lock->rw_sem);
	if (atomic_read(&Lock->in_use) && (ip->i_jfs_flag & IDIRTY)) {
		mark_inode_dirty(ip);
		ip->i_jfs_flag &= ~IDIRTY;
	}
}
#endif				/* kern22 */

#define IREAD_LOCK(ip)		READ_LOCK(&(ip)->i_jfs_rdwrlock)
#define IREAD_UNLOCK(ip)	READ_UNLOCK(ip, &(ip)->i_jfs_rdwrlock)
#define IWRITE_LOCK(ip)		WRITE_LOCK(&(ip)->i_jfs_rdwrlock)
#define IWRITE_TRYLOCK(ip)	WRITE_TRYLOCK(&(ip)->i_jfs_rdwrlock)
#define IWRITE_UNLOCK(ip)	WRITE_UNLOCK(ip, &(ip)->i_jfs_rdwrlock)
#define IWRITE_LOCK_LIST	iwritelocklist

void iwritelocklist(int, ...);

#endif				/* _H_JFS_LOCK */
